@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 # vi:si:et:sw=4:sts=4:ts=4
-# GPL 2008-2012
+# GPL 2008-2014
 
 from __future__ import division, with_statement
 
@@ -8,196 +8,12 @@ from glob import glob
 import Image
 import math
 import os
-from time import time, strftime
-
-import gobject
-import gst
-
-from imagesink import ImageSink
+from time import time
 import ox
-
+from .ffmpeg import Video
 
 FPS = 25
 
-
-class Video(gst.Pipeline):
-
-    def __init__(self, uri, height, audio, video_callback, done_callback):
-
-        gst.Pipeline.__init__(self)
-        self.duration = -1
-
-        self.height = height
-        self.video = self.height > 0
-        self.audio = audio
-
-        self.video_callback = video_callback
-        self.done_callback = done_callback
-        self.framerate = gst.Fraction(FPS, 1)
-        self.ready = False
-
-        self.src = gst.element_factory_make('filesrc')
-        self.src.props.location = uri
-        self.sbin = gst.element_factory_make('decodebin2')
-        #self.sbin.props.caps = gst.Caps('video/x-raw-yuv;video/x-raw-rgb')
-        self.sbin.props.expose_all_streams = True
-        self.add(self.src, self.sbin)
-
-        info = ox.avinfo(uri)
-        self.audio = self.audio and info['audio'] != []
-        self.video = self.video and info['video'] != []
-
-        if self.video:
-            ratio = info['video'][0]['width'] / info['video'][0]['height']
-            self.width = int(round(self.height * ratio))
-            if self.width % 4:
-                self.width += 4 - self.width % 4
-            self.vqueue = gst.element_factory_make('queue')
-            self.scale = gst.element_factory_make('videoscale')
-            self.rate = gst.element_factory_make('videorate')
-            self.csp = gst.element_factory_make('ffmpegcolorspace')
-            self.vsink = ImageSink(self._video_callback)
-            #self.vsink.connect('frame', self._video_callback)
-            self.add(
-                self.vqueue, self.scale, self.rate, self.csp, self.vsink
-            )
-
-        if self.audio:
-            self.volume = []
-            self.channels = 2
-            self.aqueue = gst.element_factory_make('queue')
-            self.convert = gst.element_factory_make('audioconvert')
-            self.resample = gst.element_factory_make('audioresample')
-            self.level = gst.element_factory_make('level')
-            # * 0.1 makes the interval about 23 msec, as opposed to about 46 msec otherwise
-            self.level.props.interval = int(gst.SECOND / float(self.framerate) * 0.1)
-            self.asink = gst.element_factory_make('fakesink')
-            self.add(
-                self.aqueue, self.convert, self.resample, self.level, self.asink
-            )
-            self.nanoseconds_per_frame = FPS / 1e9
-
-        self.src.link(self.sbin)
-        self.sbin.connect('pad-added', self._pad_added_callback)
-
-        self.set_state(gst.STATE_PAUSED)
-        self.get_state()
-        self.get_duration()
-        #self.frames = int(float(self.duration) * float(self.framerate) / gst.SECOND)
-
-        bus = self.get_bus()
-        bus.add_signal_watch()
-        self.watch_id = bus.connect('message', self._bus_message_callback)
-        self.mainloop = gobject.MainLoop()
-
-    def _pad_added_callback(self, sbin, pad):
-        caps = pad.get_caps()
-        if self.height and 'video' in str(caps):
-            pad.link(self.vqueue.get_pad('sink'))
-            self.vqueue.link(self.scale)
-            self.scale.link(self.rate)
-            self.rate.link(self.csp, gst.Caps(
-                'video/x-raw-rgb;video/x-raw-yuv,framerate=%s/%s,width=%s,height=%s' % (
-                    self.framerate.num, self.framerate.denom, self.width, self.height
-                )
-            ))
-            self.csp.link(self.vsink)
-        if self.audio and 'audio' in str(caps):
-            self.samplerate = caps[0]['rate']
-            if not isinstance(self.samplerate, int):
-                self.samplerate = self.samplerate.high
-            pad.link(self.aqueue.get_pad('sink'))
-            self.aqueue.link(self.convert)
-            self.convert.link(self.resample, gst.Caps(
-                'audio/x-raw-int,channels=%s,width=16,depth=16' % self.channels
-            ))
-            self.resample.link(self.level, gst.Caps(
-                'audio/x-raw-int,rate=%s,channels=%s,width=16,depth=16' % (
-                    self.samplerate, self.channels
-                )
-            ))
-            self.level.link(self.asink)
-
-    def _video_callback(self, frame_image, timestamp):
-        if not self.ready:
-            self.ready = True
-        else:
-            if self._is_between_in_and_out(timestamp, 'video'):
-                self.video_callback(frame_image, timestamp)
-
-    def _bus_message_callback(self, bus, message):
-        if self.audio and message.src == self.level:
-            struct = message.structure
-            if struct.get_name() == 'level':
-                timestamp = struct['timestamp']
-                if self._is_between_in_and_out(timestamp, 'audio'):
-                    sample_i = timestamp * self.nanoseconds_per_frame
-                    if sample_i > len(self.volume):
-                        self.volume.append((
-                            pow(10, struct['rms'][0] / 20),
-                            pow(10, struct['rms'][1] / 20)
-                        ))
-        elif message.src == self and message.type == gst.MESSAGE_EOS:
-            self._quit()
-
-    def _is_between_in_and_out(self, timestamp, av):
-        try:
-            if timestamp < self.in_time:
-                return False
-            if timestamp >= self.out_time:
-                self.done[av] = True
-                if self.done['video'] and self.done['audio']:
-                    self._quit()
-                    # gobject.idle_add(self._done)
-                return False
-            return True
-        except:
-            # weirdness:
-            # the first two times audio calls this, the timestamp is
-            # 23000000. the second time, self.in_time does not exist.
-            return False
-
-    def _quit(self):
-        if self.is_running:
-            self.is_running = False
-            self.mainloop.quit()
-
-    def decode(self, points=None):
-        if points:
-            self.in_time = points[0] * 1e9
-            self.out_time = points[1] * 1e9
-            if self.in_time > 5 * 1e9:
-                self.seek(
-                    1.0, gst.FORMAT_TIME,
-                    gst.SEEK_FLAG_FLUSH | gst.SEEK_FLAG_ACCURATE,
-                    gst.SEEK_TYPE_SET, self.in_time - 5 * 1e9,
-                    gst.SEEK_TYPE_NONE, -1
-                )
-        else:
-            self.in_time = 0
-            self.out_time = self.duration + 1
-        self.done = {'video': not self.video, 'audio': not self.audio}
-        self.set_state(gst.STATE_PLAYING)
-        self.is_running = True
-        self.mainloop.run()
-        self.done_callback(self.volume if self.audio else [])
-
-    def get_duration(self):
-        if self.duration < 0:
-            if self.video:
-                pads = self.vsink.sink_pads()
-            else:
-                pads = self.asink.sink_pads()
-            q = gst.query_new_duration(gst.FORMAT_TIME)
-            for pad in pads:
-                if pad.get_peer() and pad.get_peer().query(q):
-                    format, self.duration = q.parse_duration()
-        return self.duration
-
-    def get_size(self):
-        return (self.width, self.height)
-
-
 class Timelines():
 
     def __init__(
@@ -229,9 +45,12 @@ class Timelines():
             self.detect_cuts = True
             self.cuts = [0]
             self.no_cuts = []
+            self.short_cut = None
             self.cut_frames = []
             self.max_cut_len = 15000 # 10 minutes
             self.max_distance = 64 * math.sqrt(3 * pow(255, 2))
+        else:
+            self.cuts = []
 
         self.full_tile_w = 1920
         self.large_tile_w = 1500
@@ -252,7 +71,7 @@ class Timelines():
         self.log = log
         if log:
             self.profiler = Profiler()
-            self.profiler.set_task('gst')
+            self.profiler.set_task('ffmpeg')
 
         ox.makedirs(self.tile_path)
 
@@ -273,7 +92,7 @@ class Timelines():
                 video_file, self.large_tile_h if self.render_video else 0,
                 self.render_audio, self._video_callback, self._done_callback
             )
-            duration = video.get_duration() / 1e9
+            duration = video.get_duration()
             points = None
             if self.points:
                 in_point = None
@@ -289,6 +108,13 @@ class Timelines():
             self.file_frame_n.append(int(duration * FPS))
             self.file_points.append(points)
             offset += duration
+            #input did not have any video, dont try to render video timelines
+            if self.render_video and not video.video:
+                self.modes = ['audio']
+                self.render_antialias = \
+                self.render_slitscan = \
+                self.render_keyframes = \
+                self.render_video = False
 
         if not self.points:
             self.in_frame = 0
@@ -309,7 +135,7 @@ class Timelines():
             self.small_tile_n = int(math.ceil(self.duration / self.small_tile_w))
             self.small_tile_last_w =int(math.ceil(self.duration % self.small_tile_w)) or self.small_tile_w
 
-        if self.render_antialias:
+        if self.render_antialias or self.render_audio:
             self._open_full_tile()
         if self.render_keyframes:
             self.large_keyframes_tile_i = -1
@@ -319,10 +145,12 @@ class Timelines():
 
         self.frame_i = 0
         self.frame_offset = 0
+        self.log and self.profiler.set_task('ffmpeg')
         self.videos[0].decode(self.file_points[0])
 
         # remove tiles from previous run
         if not self.points:
+            #for mode in filter(lambda m: m!= 'audio', self.modes):
             for mode in self.modes:
                 type = 'png' if mode == 'data' else 'jpg'
                 tiles = glob('%s/timeline%s%dp*.%s' % (self.tile_path, mode, self.large_tile_h, type))
@@ -337,8 +165,8 @@ class Timelines():
 
         self.log and self.profiler.set_task('_video_callback()')
         '''
-        if timestamp != None and self.frame_i != int(round(timestamp / 1e9 * FPS)):
-            print 'WARNING: i is', self.frame_i, 'timestamp is', timestamp, '(', int(round(timestamp / 1e9 * FPS)), ')'
+        if timestamp != None and self.frame_i != int(round(timestamp * FPS)):
+            print 'WARNING: i is', self.frame_i, 'timestamp is', timestamp, '(', int(round(timestamp * FPS)), ')'
         '''
         self.is_last_frame = self.frame_i == self.frame_n - 1
         large_tile_x = self.frame_i % self.large_tile_w
@@ -387,10 +215,17 @@ class Timelines():
                 is_cut = False
                 self.previous_distance = 0
             else:
+                if self.short_cut and self.frame_i - self.short_cut > 2:
+                    self.cuts.append(self.short_cut)
+                    self.short_cut = None
                 distance = self._get_distance(self.previous_frame_data, frame_data)
                 is_cut = distance > 0.08 or (distance > 0.04 and abs(distance - self.previous_distance) > 0.04)
                 if is_cut:
-                    self.cuts.append(self.frame_i)
+                    if self.frame_i - (0 if not self.cuts else self.short_cut or self.cuts[-1]) < 3:
+                        is_cut = False
+                        self.short_cut = self.frame_i
+                    else:
+                        self.cuts.append(self.frame_i)
                 elif len(self.cut_frames) == self.max_cut_len:
                     is_cut = True
                     self.cuts.append(self.frame_i)
@@ -423,14 +258,9 @@ class Timelines():
             for mode in filter(lambda x: x in ['antialias', 'slitscan', 'cuts', 'data'], self.modes):
                 self._save_tile(mode, self.large_tile_i)
             if self.render_antialias:
-                resized = self.large_tile_image['antialias'].resize((
-                    self.full_tile_widths[0], self.large_tile_h
-                ), Image.ANTIALIAS)
-                self.full_tile_image.paste(resized, (self.full_tile_offset, 0))
-                self.full_tile_offset += self.full_tile_widths[0]
-                self.full_tile_widths = self.full_tile_widths[1:]
+                self._save_full_tile_image('antialias')
         self.frame_i += 1
-        self.log and self.profiler.set_task('gst')
+        self.log and self.profiler.set_task('ffmpeg')
 
     def _render_keyframes(self):
         self.log and self.profiler.set_task('_render_keyframes() # keyframes timelines')
@@ -441,7 +271,7 @@ class Timelines():
             frame_ratio = self.frame_ratio * wide
             cut_images = int(math.ceil(cut_width / (frame_ratio * self.large_tile_h)))
             if cut_images == 0:
-                print 'ERROR division by zerro', cut_width, cut_images
+                print 'ERROR division by zero', cut_width, cut_images
                 index = -1
             image_widths = self._divide(cut_width, cut_images)
             image_i = self.cuts[-2]
@@ -510,6 +340,10 @@ class Timelines():
             is_last_tile = self.frame_i == self.frame_n - 1
             if large_tile_x == self.large_tile_w - 1 or is_last_tile:
                 self._save_tile('audio', large_tile_i)
+                if not self.render_video:
+                    self._save_full_tile_image('audio')
+        if not self.render_video:
+            self.large_tile_i = large_tile_i
 
     def _divide(self, num, by):
         # divide(100, 3) -> [33, 33, 34]
@@ -537,7 +371,7 @@ class Timelines():
 
     def _open_full_tile(self):
         if self.large_tile_n == 1:
-            self.full_tile_widths = [self.large_tile_last_w]
+            self.full_tile_widths = [self.full_tile_w]
         else:
             full_tile_w = self.full_tile_w
             full_tile_n = self.large_tile_n
@@ -552,10 +386,18 @@ class Timelines():
         self.full_tile_offset = 0
         self.full_tile_image = Image.new('RGB', (self.full_tile_w, self.large_tile_h))
 
-    def _save_full_tile(self):
+    def _save_full_tile_image(self, mode):
+        resized = self.large_tile_image[mode].resize((
+            self.full_tile_widths[0], self.large_tile_h
+        ), Image.ANTIALIAS)
+        self.full_tile_image.paste(resized, (self.full_tile_offset, 0))
+        self.full_tile_offset += self.full_tile_widths[0]
+        self.full_tile_widths = self.full_tile_widths[1:]
+
+    def _save_full_tile(self, mode):
         self.log and self.profiler.set_task('_save_full_tile()')
-        tile_file = '%stimelineantialias%dp.jpg' % (
-            self.tile_path, self.large_tile_h
+        tile_file = '%stimeline%s%dp.jpg' % (
+            self.tile_path, mode, self.large_tile_h
         )
         self.full_tile_image.save(tile_file)
         if self.log:
@@ -563,8 +405,8 @@ class Timelines():
         if self.render_small_tiles:
             resize = (self.full_tile_w, self.small_tile_h)
             self.full_tile_image = self.full_tile_image.resize(resize, Image.ANTIALIAS)
-            tile_file = '%stimelineantialias%dp.jpg' % (
-                self.tile_path, self.small_tile_h
+            tile_file = '%stimeline%s%dp.jpg' % (
+                self.tile_path, mode, self.small_tile_h
             )
             self.full_tile_image.save(tile_file)
             if self.log:
@@ -633,20 +475,22 @@ class Timelines():
         if self.video_i < self.video_n - 1:
             self.video_i += 1
             self.frame_offset = self.frame_i
+            self.log and self.profiler.set_task('ffmpeg')
             self.videos[self.video_i].decode(self.file_points[self.video_i])
         else:
             if self.render_video:
                 if self.render_antialias:
-                    self._save_full_tile()
+                    self._save_full_tile('antialias')
                 if self.cuts_path:
-                    # remove cuts at max_cut_len
-                    self.cuts = filter(lambda x: x not in self.no_cuts, self.cuts)
+                    # remove cuts at first and last frame and all cuts at max_cut_len
+                    self.cuts = filter(lambda x: x not in [0, self.frame_n] + self.no_cuts, self.cuts)
                     with open(self.cuts_path, 'w') as f:
-                        # remove 0 and frame_n from cuts
                         # avoid float rounding artefacts
-                        f.write('[' + ', '.join(map(lambda x: '%.2f' % (x / FPS), self.cuts[1:-1])) + ']')
+                        f.write('[' + ', '.join(map(lambda x: '%.2f' % (x / FPS), self.cuts)) + ']')
             if self.render_audio:
                 self._render_audio(self.volume)
+                if not self.render_video:
+                    self._save_full_tile('audio')
             if self.log:
                 if self.render_video:
                     maximum = 0
